from .enricher.elasticsearch_connector import (
    ElasticsearchConnector,
    IndexHandler,
    ReindexHandler,
    PipelineHandler,
    PolicyHandler,
)

from .enricher.osm.osm_places_connector import OSMPlacesConnector
from .enricher.osm.functional_region_connector import FunctionalRegionConnector

from .enricher.api_connector import ApiConnector
from .enricher.sql_connector import SqlConnector
from .enricher import Enricher, EnricherBuilder, IEnricherConnector

from .models import Data
from .parser import parse_document
from .utils import write_json_generator_to_json, Converter

__all__ = [
    "ElasticsearchConnector",
    "IndexHandler",
    "ReindexHandler",
    "PipelineHandler",
    "PolicyHandler",
    "ApiConnector",
    "OSMPlacesConnector",
    "FunctionalRegionConnector",
    "Sqlconnector",
    "Enricher",
    "EnricherBuilder",
    "IEnricherConnector",
    "Data",
    "parse_document",
    "write_json_generator_to_json",
    "Converter",
]



from ..enricher import IEnricherConnector

import pandas as pd
import os
from shapely import wkt
import geopandas

from shapely.geometry import MultiPolygon, Polygon
from shapely.ops import transform
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from shapely.geometry import shape, mapping
from rtree import index as rtreeindex
from moredata.models.data import GeopandasData, JsonData, DaskGeopandasData
import dask_geopandas


class FunctionalRegionConnector(IEnricherConnector):
    """FunctionalRegionConnector implements IEnricherConnector interface, so this is a connector that can be used to enrich data. This connector counts how much elements is within a radius of a lat/lon specified by your data.

    Parameters
    ----------
    files: List[str]
        An array of files that will be used to enrich your data. Each file must have 1 attribute, an geometry specifying the polygon to index in RTree.

    dict_keys: List[str]
        The connector doesn't know where to get data to make the relationship, so you have to pass the keys or if your data isn't nested, just one key to the connector reach at the the right attribute.

    key: str
        Key attribute is the name of the new column of your data that stores the counting.

    Attributes
    ----------
    files: List[str]

    dict_keys: List[str]

    key: str
    """

    def __init__(self, files, key, dict_keys=[]):
        self.key = key
        self.dict_keys = dict_keys
        self.files = files

        self._df = []
        if self.files is not None:
            read_temp = []
            for file in self.files:
                read_temp.append(pd.read_csv(file))
            self._df = pd.concat(read_temp)
            self._df["geometry"] = self._df["geometry"].apply(wkt.loads)
            self._df = geopandas.GeoDataFrame(self._df)

    def _get_polygons(self):
        self.idx = rtreeindex.Index()

        for f in self.files:
            array_polygons = []
            for index, row in self._df.iterrows():
                pol = row["geometry"]
                if isinstance(pol, Polygon) or isinstance(pol, Point):
                    array_polygons.append(pol)

            for pos, poly in enumerate(array_polygons):
                self.idx.insert(pos, poly.bounds)

    def _fence_check_local(self, point):
        count = 0

        shp = wkt.loads(point["area_point"])
        for j in self.idx.intersection(shp.bounds):
            count += 1
        return count

    def _enrich_point(self, point):
        if "area_point" in point.keys():
            amount = self._fence_check_local(point)

            if not self.key in point.keys():
                point[self.key] = amount
            else:
                point[self.key] += amount
        else:
            raise Exception(
                "area_point polygon was not found. Please use function geodesic_point_buffer present in utils package and try again."
            )

    def enrichJsonData(self, data, **kwargs):
        for d in data.parse(**kwargs):

            if not self.dict_keys:
                points = d
            else:
                points = d[self.dict_keys[0]]
                for k in range(1, len(self.dict_keys)):
                    try:
                        points = points[self.dict_keys[k]]
                    except KeyError as e:
                        return None

            if isinstance(points, list):
                for point in points:
                    self._enrich_point(point)
            else:
                self._enrich_point(points)

            yield d

    def enrich_geopandas_data(self, data):
        self._df = self._df.set_crs(epsg=4326).to_crs(epsg=3857)
        data.reset_index(inplace=True)

        spatial_joined = geopandas.sjoin(
            data, self._df, how="inner", predicate="intersects"
        )
        different_indices = spatial_joined.index.value_counts()

        data.set_index("index", inplace=True)

        data[self.key] = different_indices
        data[self.key].fillna(0, inplace=True)

        return GeopandasData.from_geodataframe(data)

    def enrich(self, data, **kwargs):
        """Method overrided of interface. It walk through the keys to reach at the data that will be used to intersect the polygons. It uses a R tree to index polygons and search faster. For optimization purposes we recommend to buffer the point, using ``geodesic_point_buffer`` function, creating, necessarily, a label named ``area_point``, and save the file with points buffered to use as base of enrichment. After buffer the points, you can use the Functional Region Connector to create your enrichment passing proper attributes."""
        self._get_polygons()

        if isinstance(data, GeopandasData):
            return self.enrich_geopandas_data(data.data)

        elif isinstance(data, DaskGeopandasData):
            raise Exception("DaskGeopandasData not supported yet")

        elif isinstance(data, JsonData):
            return self.enrichJsonData(data, **kwargs)



import os

__all__ = ["available", "get_path"]

_module_path = os.path.dirname(__file__)
_available_csv = {
    "airbnb-berlin-main": "./airbnb/airbnb-berlin-main.csv",
    "airbnb-berlin-host": "./airbnb/airbnb-berlin-host.csv",
    "airbnb-berlin-extra": "./airbnb/airbnb-berlin-extra.csv",
}
available = list(_available_csv.keys())


def get_path(dataset):
    if dataset in _available_csv:
        return os.path.abspath(os.path.join(_module_path, _available_csv[dataset]))
    else:
        msg = "The dataset '{data}' is not available. ".format(data=dataset)
        msg += "Available datasets are {}".format(", ".join(available))
        raise ValueError(msg)



from elasticsearch.client import IngestClient
from json import dump
from ...utils import load_json, read_json_from_file


class PipelineHandler:
    """
    PipelineHandler create json object for elasticsearch pipeline.

    Parameters
    ----------
    description: str
        explain what is pipeline.
    target_field_name: str
        field name that will post the result of enrichment.
    match_field: str
        what index lookup field to bind the enrichments.
    policy_name: str
        name of policy.
    field_array: str
        name of field if the field you want to enrich is an array.
    max_matches: int
        if the matches is > 1 the target_field will be an array and
        the results of enrichments will bind more than 1 object to get result.
    shape_relation: str
        if the enrichment is geobased you have to put this parameter to specify
        which relation you want to get.

        - INTERSECTS
            Return all documents whose field intersects the query geometry.
        - DISJOINT
            Return all documents whose field has nothing in common with the query geometry.
        - WITHIN
            Return all documents whose field is within the query geometry.
        - CONTAINS
            Return all documents whose field contains the query geometry.

    Attributes
    ----------

    json: dict
        this attribute is the json created with the parameters specifying the pipeline.
    """

    def __init__(
        self, description, target_field_name, match_field, policy_name, **kwargs
    ):

        if kwargs.get("field_array"):
            self._json = {
                "description": description,
                "processors": [
                    {
                        "foreach": {
                            "field": kwargs.get("field_array"),
                            "processor": {
                                "enrich": {
                                    "policy_name": policy_name,
                                    "field": "_ingest._value." + match_field,
                                    "target_field": "_ingest._value."
                                    + target_field_name,
                                    "ignore_missing": True,
                                }
                            },
                            "ignore_failure": True,
                        }
                    }
                ],
            }

            if kwargs.get("remove_field"):
                remove_processor = {
                    "foreach": {
                        "field": kwargs.get("field_array"),
                        "processor": {
                            "remove": {
                                "field": "_ingest._value." + kwargs.get("remove_field"),
                                "ignore_missing": True,
                            }
                        },
                        "ignore_failure": True,
                    }
                }
                self._json["processors"].append(remove_processor)

            if kwargs.get("max_matches"):
                self._json["processors"][0]["foreach"]["processor"]["enrich"][
                    "max_matches"
                ] = kwargs.get("max_matches")

            if kwargs.get("shape_relation"):
                self._json["processors"][0]["foreach"]["processor"]["enrich"][
                    "shape_relation"
                ] = kwargs.get("shape_relation")
        else:
            self._json = {
                "description": description,
                "processors": [
                    {
                        "enrich": {
                            "policy_name": policy_name,
                            "field": match_field,
                            "target_field": target_field_name,
                            "ignore_missing": True,
                        }
                    }
                ],
            }

            if kwargs.get("remove_field"):
                remove_processor = {
                    "remove": {
                        "field": kwargs.get("remove_field"),
                        "ignore_missing": True,
                    }
                }

                self._json["processors"].append(remove_processor)

            if kwargs.get("max_matches"):
                self._json["processors"][0]["enrich"]["max_matches"] = kwargs.get(
                    "max_matches"
                )

            if kwargs.get("shape_relation"):
                self._json["processors"][0]["enrich"]["shape_relation"] = kwargs.get(
                    "shape_relation"
                )

    def import_json(self, filename):
        self._json = read_json_from_file(filename)

    def export_json(self, filename):
        with open(filename, "w", encoding="utf-8") as f:
            dump(self._json, f, ensure_ascii=False, indent=4)


class Pipeline:
    """
    A pipeline is a definition of a series of processors that
    are to be executed in the same order as they are declared.
    (https://www.elastic.co/guide/en/elasticsearch/reference/current/pipeline.html)

    Parameters
    ----------

    client: elasticsearch.Elasticsearch
        a elasticsearch client.
    name: str
        name for pipeline.
    pipeline_handler: :obj:`PipelineHandler`
        object that contains json object of pipeline.

    Attributes
    ----------

    client: str
    name: str
    pipeline_handler: :obj:`PipelineHandler`
    """

    def __init__(self, client, name, pipeline_handler):
        self._pipeline_handler = pipeline_handler
        self._name = name
        self.ingest_client = IngestClient(client)

    def create_pipeline(self, params=None):
        """
        create_pipeline method create the elasticsearch pipeline
        with processors specified in pipeline_handler json.
        """
        try:
            self.ingest_client.put_pipeline(
                self._name, load_json(self._pipeline_handler._json)
            )
        except Exception as e:
            raise (e)



import geopandas
import dask_geopandas


class Data:
    """Data object represents a file like csv, json
    with a function to parser that file.

    Parameters
    ----------
    parser_func: function
    data_type: str
        type of data, the extension of file.
    data_file: str
        where the data is.


    Attributes
    ----------

    parser_func: function

    data_type: str

    data_file: str

    enrichers: array
        array of enrichers.
    """

    def __init__(self):
        self.enrichers = []

    def add(self, enricher):
        """add enricher in enrichers attribute"""
        self.enrichers.append(enricher)


class GeopandasData(Data):
    @classmethod
    def from_geodataframe(cls, geodataframe):
        geopandas_data = GeopandasData()
        geopandas_data.data = geodataframe
        return geopandas_data

    @classmethod
    def from_path(cls, path):
        geopandas_data = GeopandasData()
        geopandas_data.data = geopandas.read_file(path)
        return geopandas_data


class DaskGeopandasData(Data):
    @classmethod
    def from_geodataframe(cls, geodataframe, npartitions=4):
        dgd = DaskGeopandasData()
        dgd.data = dask_geopandas.from_geopandas(geodataframe, npartitions)
        return dgd

    @classmethod
    def from_path(cls, path, npartitions=4):
        dgd = DaskGeopandasData()
        dgd.data = dask_geopandas.read_file(path, npartitions)
        return dgd


class JsonData(Data):
    def __init__(self, data_file, parser):
        self.data_file = data_file
        self.parser = parser

    def parse(self, **kwargs):
        """parse calls the parser_func attribute"""
        return self.parser(self.data_file, **kwargs)



from ..enricher import IEnricherConnector
from ...models.data import GeopandasData, JsonData
from ...utils import load_json


class ApiConnector(IEnricherConnector):
    """ApiConnector implements interface IEnricherConnector,
    so this connector can be used to enrich some data.

    Parameters
    ----------
    response_parser: Callable
        a function that will parse the response of the request
        and returns a dict with the values which will enrich the data.

    url_pattern: str
        url_pattern is the route of the api to retrieve values
        which will enrich the data.

    params: Dict
        params are a dict that has key as the variable in route and name is
        where to retrieve the data value to assign in route.

    Attributes
    ----------
    response_parser: Callable

    url_pattern: str

    params: dict
    """

    def __init__(self, response_parser, url_pattern, params):
        self.response_parser = response_parser
        self.url_pattern = url_pattern
        self.params = params

    def _handle_params(cls, pattern, params):
        """
        Replace all variables inside pattern string with params
        values to build the url to request.

        Parameters
        ----------
        pattern: str

        params: Dict

        Returns
        -------
        url: str
        """
        import re

        regex = re.compile("\{.*?\}")
        routes = regex.findall(pattern)

        for route in routes:
            pattern = pattern.replace(route, str(params[route[1 : len(route) - 1]]))

        return pattern

    def _make_request(cls, url, *args, **kwargs):
        """
        Using requests package this method will do the request and return that json.

        Parameters
        ----------
        url: str

        Returns
        -------
        json: Json
        """
        import requests
        from urllib3.exceptions import InsecureRequestWarning

        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        return requests.get(url, verify=False).json()

    def enrichJsonData(self, data, **kwargs):
        responses_cache = {}

        for d in data.parse(**kwargs):
            if self.params["fields"]:
                for field in self.params["fields"]:
                    self.params[field["key"]] = d[field["name"]]

                url = self._handle_params(self.url_pattern, self.params)
                if url in responses_cache.keys():
                    response_value = responses_cache[url]
                else:
                    response_value = self.response_parser(self._make_request(url))
                    responses_cache[url] = response_value

                if response_value is not None:
                    for k, v in response_value.items():
                        d[k] = v

                yield d

    def enrichGeoPandasData(self, data):
        responses_cache = {}

        d = data.copy()
        for i, row in d.iterrows():
            if self.params["fields"]:

                for field in self.params["fields"]:
                    self.params[field["key"]] = row[field["name"]]
                url = self._handle_params(self.url_pattern, self.params)

                if url in responses_cache.keys():
                    response_value = responses_cache[url]
                else:
                    response_value = self.response_parser(self._make_request(url))
                    responses_cache[url] = response_value

                if response_value is not None:
                    for k, v in response_value.items():
                        d.loc[i, k] = v
        return d

    def enrich(self, data, **kwargs):
        """Method overrided of interface. This interface do enrichment using
        API as a enricher and return all data enriched as Json. This method
        has a cache that will store as dict the url as key and the response parsed as value
        for don't spend time with requests that are already done previously.

        Parameters
        ----------
        data: Data

        Yields
        ------
        data: Dict
        """

        if isinstance(data, GeopandasData):
            return self.enrichGeoPandasData(data.data)

        elif isinstance(data, JsonData):
            return self.enrichJsonData(data, **kwargs)



from moredata.datasets import get_path
from geopandas import GeoDataFrame, points_from_xy
import pandas as pd

import pytest


@pytest.mark.parametrize("test_dataset", ["airbnb-berlin"])
def test_read_paths(test_dataset):
    df = pd.read_csv(get_path(test_dataset), low_memory=False)
    assert isinstance(
        GeoDataFrame(df, geometry=points_from_xy(df.latitude, df.longitude)),
        GeoDataFrame,
    )



from elasticsearch.client.enrich import EnrichClient
from json import dump
from ...utils import load_json, read_json_from_file


class PolicyHandler:
    """PolicyHandler creates json object for elasticsearch Policy

    Parameters
    ----------
    type_match: str
            - geo_match
                    enrich data to incoming documents based on a geographic
                    location using a geo_shape query.
            - match
                    enrich data to incoming documents based on a precise value,
                    such as an email address or ID, using a term query.

    index: str
            name of source index.

    match_field: str
            field in the source indices used to match incoming documents.

    enrich_fields: array
            fields to add to matching incoming documents.
            These fields must be present in the source indices
    """

    def __init__(self, type_match, index, match_field, enrich_fields):
        self._json = {
            type_match: {
                "indices": index,
                "match_field": match_field,
                "enrich_fields": enrich_fields,
            }
        }

    def import_json(self, filename):
        self._json = read_json_from_file(filename)

    def export_json(self, filename):
        with open(filename, "w", encoding="utf-8") as f:
            dump(self._json, f, ensure_ascii=False, indent=4)


class Policy:
    """
    A set of configuration options used to add the right
    enrich data to the right incoming documents
    (https://www.elastic.co/guide/en/elasticsearch/reference/current/enrich-policy-definition.html).

    Parameters
    ----------

    client: elasticsearch.Elasticsearch
    a elasticsearch client.
    policy_handler: :obj:`PolicyHandler`
            a policy handler that create json policy
    name: str
            name for policy enrichment

    Attribute
    ----------

    client: elasticsearch.Elasticsearch
    policy_handler: :obj:`PolicyHandler`
    name: str
    """

    def __init__(self, client, policy_handler, name):

        self._policy_handler = policy_handler
        self.name = name
        self.enrich_client = EnrichClient(client)

    def create_policy(self, params=None):
        """create_policy is a method to create policy file"""
        try:
            self.enrich_client.put_policy(
                self.name, load_json(self._policy_handler._json)
            )
        except Exception as e:
            raise (e)

    def execute_policy(self, params=None):
        """execute_policy create the enrich index for an existing enrich policy"""
        try:
            self.enrich_client.execute_policy(self.name)
        except Exception as e:
            raise (e)



from .osm_downloader import OSM_util

import json
import pyproj
from functools import partial
from shapely.ops import transform
from shapely.geometry import Point, Polygon


def geodesic_point_buffer(lat, lon, radius):
    """This method implements a geodesical point buffering. It creates a circle with radius around the lat/long. It uses the azhimutal projection to fix the problem of distances proportions on globe.

    Parameters
    ----------
    lat: float

    lon: float

    radius: float
        distance in meters
    """
    wgs84 = pyproj.CRS("EPSG:4326")
    merc = pyproj.CRS("EPSG:3857")
    x, y = pyproj.transform(wgs84, merc, lat, lon)
    poly = Polygon(Point(x, y).buffer(radius).exterior.coords[:])

    project = pyproj.Transformer.from_crs(merc, wgs84).transform
    poly = transform(project, poly)

    return poly


def read_json_from_file(file):
    with open(file, "r") as f:
        return json.loads(f.read())


def load_json(json_object):
    return json.loads(json.dumps(json_object))


def chunks(iterable, size=10):
    from itertools import chain, islice

    iterator = iter(iterable)
    for first in iterator:
        yield chain([first], islice(iterator, size - 1))


def write_json_generator_to_json(file, data, n):
    for i, group in enumerate(chunks(data, n)):
        with open(file + "-{}.json".format(i), "w") as outfile:
            json.dump(list(group), outfile, ensure_ascii=False)


class Converter:
    @staticmethod
    def json_enriched_to_csv(path, output_path):
        """Get the output of json and converts to csv file.

        Parameters
        ----------

        path: str
            path where json files are.

        output_path: str
            path where the conversion files will be.
        """
        import pandas as pd
        import json
        from glob import glob

        files = glob(path)

        for i, file in enumerate(files):
            try:
                df = pd.read_json(file, orient="records")
                df.to_csv(output_path + str(i) + ".csv", encoding="utf-8", index=False)
            except AttributeError as e:
                print(file)
                raise (e)

    @staticmethod
    def json_enriched_to_parquet(path, output_path):
        """Get the output of json and converts to parquet file.

        Parameters
        ----------

        path: str
            path where json files are.

        output_path: str
            path where the conversion files will be.
        """
        import pandas as pd
        from glob import glob

        files = glob(path)

        for i, file in enumerate(files):
            try:
                df = pd.read_json(file, orient="records")
                df.to_parquet(
                    output_path + str(i) + ".parquet", encoding="utf-8", index=False
                )
            except AttributeError as e:
                pass

    @staticmethod
    def csv_to_json(file, output_file):
        """Get the output of csv and converts to json file.

        Parameters
        ----------

        file: str
            name of file which you want to convert.

        output_file: str
            name of output_file.
        """
        import pandas as pd

        df = pd.read_csv(file)
        df.to_json(output_file, orient="records")

    @staticmethod
    def parquet_to_json(file, output_file):
        """Get the output of parquet and converts to json file.

        Parameters
        ----------

        file: str
            name of file which you want to convert.

        output_file: str
            name of output_file.
        """
        import pandas as pd

        df = pd.read_parquet(file)
        df.to_json(output_file, orient="records")



import json
import csv
from h3 import h3
from ..utils import read_json_from_file, load_json
from shapely.geometry import asPoint
from numpy import array


def _add_geo_location(doc):
    doc["geo_location"] = asPoint(array([doc["longitude"], doc["latitude"]])).wkt
    return doc


def _add_code_point(doc):
    doc["code_h3"] = h3.geo_to_h3(doc["latitude"], doc["longitude"], 8)
    return doc


def parse_document(data, **kwargs):
    array_point_field = kwargs.get("array_point_field")
    geo_location = kwargs.get("geo_location")
    code_h3 = kwargs.get("code_h3")

    for doc in read_json_from_file(data):
        if geo_location:
            if array_point_field != None:
                doc[array_point_field] = [
                    _add_geo_location(points) for points in doc[array_point_field]
                ]
            else:
                doc = _add_geo_location(doc)

        if code_h3:
            if array_point_field != None:
                doc[array_point_field] = [
                    _add_code_point(points) for points in doc[array_point_field]
                ]
            else:
                doc = _add_code_point(doc)
        yield doc



from abc import ABC, abstractproperty, abstractmethod
from logging import raiseExceptions

from moredata.enricher.elasticsearch_connector.elasticsearch_connector import (
    ElasticsearchConnector,
)
from .enricher import Enricher
from ..models.data import Data, JsonData


class Builder(ABC):
    """Design pattern Builder interface"""

    @abstractmethod
    def with_enrichment(self, connector):
        """This method is suppose to add enrichments to :obj:`Data` class
        Parameters
        ----------

        connector: :obj:`Enricher`
            a connector that will implement enrich method
        """
        pass


class EnricherBuilder(Builder):
    """EnricherBuilder is a class that override Builder interface and for that
    implements abstract method with_enrichment.

    Parameters
    ----------
    data: :obj:`Data`
        Data class is what will be enriched.

    Attributes
    ----------
    data: :obj:`Data`

    """

    def __init__(self, data):
        self._data = data

    def with_enrichment(self, connector):
        """Implementation of abstract method of :obj:`Builder`
        This method add to :obj:`Data` array attribute :obj:`Enricher` objects

        Parameters
        ----------
        connector: :obj:`Enricher`
            a connector that will implement :func:`~enricher.Enricher.enrich`

        Returns
        -------
        This method return the :obj:`Data`
        """

        if isinstance(self._data, JsonData) and not isinstance(
            connector, ElasticsearchConnector
        ):
            raise Exception(
                "JsonData can only be enriched with ElasticsearchConnector using EnricherBuilder"
            )

        self._data.add(connector)
        return self

    def get_result(self, **kwargs):
        def pipe(data, funcs, **kwargs):
            for func in funcs:
                data = func(data=data, **kwargs)
            return data

        """Implements a chaining method like `f(g(h(data)))`. So,
        this method call all of enrichments functions that are 
        saved as array attribute in data class. 

        Returns
        -------
        This method returns the implementations of each :func:`~enricher.Enricher.enrich`.
        """
        return pipe(self._data, [e.enrich for e in self._data.enrichers], **kwargs)



from abc import ABC, abstractmethod
from ..models.data import Data
from moredata.models.data import DaskGeopandasData


class Enricher:
    """Enricher is a wrapper class of interface IEnricherConnector

    Parameters
    ----------
    connector: :obj:`IEnricherConnector`
        Interface that connects an implementation of enrichment to the class :Enricher:

    Attributes
    ----------
    connector: :obj:`IEnricherConnector`
        Interface that connects an implementation of enrichment to the class :Enricher:

    """

    def __init__(self, connector):
        self._connector = connector

    @property
    def connector(self):
        """
        Returns
        -------
            :obj:`IEnricherConnector` getter method
        """
        return self._connector

    @connector.setter
    def connector(self, connector):
        self._connector = connector

    def enrich(self, data, **kwargs):
        """Call enrich function of interface :obj:`IEnricherConnector`

        Parameters
        ----------
        data: :obj:`Data`

        Returns
        -------
        :obj:`Data`
            data enriched

        """
        return self._connector.enrich(data, **kwargs)


class IEnricherConnector(ABC):
    """IEnricherConnector is an interface that provides the abstract method enrich."""

    @abstractmethod
    def enrich(self, data, **kwargs):
        pass



from ..enricher import IEnricherConnector
import sqlalchemy
import decimal
import pandas as pd
from moredata.models.data import GeopandasData, JsonData


class SqlConnector(IEnricherConnector):
    """SQLconnector implements interface IEnricherConnector, so this is a connector that can be used to enrich data.

    Parameters
    ----------
    table_name: str
        name of table that will be used to enrich

    dict_keys: List[str]
        dict_keys is the principal argument to the connector.
        The connector doesn't know where to get data to make the relationship,
        so you have to pass the keys or if your data isn't nested, just one key
        to the connector reach at the the right attribute.

    result_attr: str
        where tha enriched data it supposed to be

    column: str
        name of table column to build the relation.

    engine: :obj:`sqlalchemy.engine.Engine`
        an engine already created by sqlalchemy

    connection_url: str
        the string of connection with host, port, password and database.

    Attributes
    ----------
    table_name: str

    dict_keys: List

    result_attr: str

    column: str

    engine: :obj:`sqlalchemy.engine.Engine`, optional

    connection_url: str, optional
    """

    def __init__(
        self,
        table_name,
        result_attr,
        column,
        dict_keys=[],
        engine=None,
        connection_url=None,
        df_column=None,
    ):
        if engine is None and connection_url is None:
            raise Exception

        if engine is not None:
            self.engine = engine

        if connection_url is not None:
            self.engine = sqlalchemy.create_engine(connection_url)

        self.table = sqlalchemy.Table(
            table_name, sqlalchemy.MetaData(), autoload_with=self.engine
        )
        self.column = column
        self.dict_keys = dict_keys
        self.result_attr = result_attr
        self.df_column = df_column

    def _enrich_object(self, obj):
        def get_value(obj):
            query = self.table.select().where(self.table.c[self.column] == obj)
            result = self.engine.execute(query)

            first = result.first()

            if first is not None:
                result_dict = dict(first)

                if result_dict is not None:
                    for k, _ in result_dict.items():
                        if isinstance(result_dict[k], decimal.Decimal):
                            result_dict[k] = float(result_dict[k])

                return result_dict
            return None

        if isinstance(obj, list):
            result_dict = []
            for o in obj:
                result_dict.append(get_value(o))
            return result_dict

        return get_value(obj)

    def enrich_geopandas_data(self, data):
        d = data.copy()
        for i, row in d.iterrows():
            ret = self._enrich_object(row[self.df_column])
            d.loc[i, "enriched"] = str(ret)
        return d

    def enrich_json_data(self, data, **kwargs):
        for d in data.parse(**kwargs):
            objects = d[self.dict_keys[0]]
            for k in range(1, len(self.dict_keys) - 1):
                try:
                    objects = objects[self.dict_keys[k]]
                except KeyError as e:
                    return None

            if isinstance(objects, list):
                if isinstance(objects[0], dict):
                    for obj in objects:
                        obj[self.result_attr] = self._enrich_object(
                            obj[self.dict_keys[-1]]
                        )
                else:
                    d[self.result_attr] = self._enrich_object(objects)
            else:
                d[self.result_attr] = self._enrich_object(objects)

            yield d

    def enrich(self, data, **kwargs):
        """Method overrided of interface. This method do enrichment using RDBMS
        as a enricher. It walk through the keys to reach at the data that will
        be used to create the relationship. After, if the object is a list it
        creates an attribute on parent object, if the object is just a dict it
        creates an attribute inside.

        Parameters
        ----------
        data: :obj:`Data`
        """

        if isinstance(data, GeopandasData):
            if self.df_column == None:
                raise Exception("df_column is required in GeopandasData")
            return self.enrich_geopandas_data(data.data)

        elif isinstance(data, JsonData):
            return self.enrich_json_data(data, **kwargs)



import dask_geopandas
from moredata.models.data import GeopandasData, JsonData, DaskGeopandasData
from ..enricher import IEnricherConnector
from ...utils import OSM_util
import pandas as pd
from shapely import wkt
import geopandas
from shapely.geometry import Point
from shapely.geometry import shape
from rtree import index as rtreeindex

from ...utils import geodesic_point_buffer


class OSMPlacesConnector(IEnricherConnector):
    """OSMconnector implements interface IEnricherConnector, so this is a connector that can be used to enrich data.

    Parameters
    ----------
    dict_keys: List[str]
        dict_keys is the principal argument to the connector. The connector doesn't know where to get data to make the relationship, so you have to pass the keys or if your data isn't nested, just one key to the connector reach at the the right attribute.

    key: str
        class of OSM. eg.: 'amenity', 'leisure'.

    value: str
        value of location of OSM. eg.: 'hospital', 'stadium'.

    place_name: str
        name of local. eg.: 'Brasil', 'São Paulo'.

    file: str, optional
        name of file in CSV of downloaded polygons with must columns: key, value, polygon.

    radius: numeric, optional
        radius to around of point to intersect the polygon.

    buffered: boolean
        True if the region is already buffered;
        False if you want buffer the region;

    files: List[str]

    Attributes
    ----------
    dict_keys: List[str]

    key: str

    value: str

    place_name: str

    file: str, optional

    radius: numeric, optional
    """

    def __init__(
        self,
        key=None,
        value=None,
        dict_keys=[],
        place_name="Brasil",
        files=None,
        radius=None,
        geometry_intersected=False,
        buffered=False,
    ):
        self.key = key
        self.value = value
        self.place_name = place_name
        self.files = files
        self.radius = radius
        self.dict_keys = dict_keys
        self.geometry = geometry_intersected
        self.buffered = buffered

        if self.files is not None:
            read_temp = []
            for file in self.files:
                read_temp.append(pd.read_csv(file))
            self._df = pd.concat(read_temp)
            self._df["geometry"] = self._df["geometry"].apply(wkt.loads)
            self._df = geopandas.GeoDataFrame(self._df)
        if self.geometry:
            self._df["geometry_intersected"] = self._df.geometry

    def _get_polygons(self):
        self.array_polygons = []
        for index, row in self._df.iterrows():
            pol = row["geometry"]
            self.array_polygons.append(pol)

        self.idx = rtreeindex.Index()
        for pos, poly in enumerate(self.array_polygons):
            self.idx.insert(pos, poly.bounds)

    def _fence_check_local(self, point):
        polygon_metadata = []
        if self.buffered:
            shp = wkt.loads(point["area_point"])
        elif self.radius is not None:
            shp = geodesic_point_buffer(
                point["latitude"], point["longitude"], self.radius
            )

        else:
            shp = Point(point["longitude"], point["latitude"])

        for j in self.idx.intersection(shp.bounds):
            if self.radius is None:
                if shp.within(shape(self.array_polygons[j])):
                    polygon_metadata.append(self._df.iloc[j].to_frame().T)
            else:
                polygon_metadata.append(self._df.iloc[j].to_frame().T)
        return polygon_metadata

    def _traverse_dict(self, dict, keys):
        for k in keys:
            try:
                dict = dict[k]
            except KeyError as e:
                return None
        return dict

    def _enrich_point(self, point):
        if "latitude" in point.keys() and "longitude" in point.keys():
            polygon_metadata = self._fence_check_local(point)

            for p in polygon_metadata:
                if not "local" in point.keys():
                    point["local"] = []
                if not "geometry_intersected" in point.keys() and self.geometry:
                    point["geometry_intersected"] = []

                if self.geometry:
                    polygons_intersected = list(p["geometry"])
                    for polygon in polygons_intersected:
                        point["geometry_intersected"].append(str(polygon))

                point["local"].append(*p[["name", "key", "value"]].to_dict("records"))

    def enrich_json_data(self, data, **kwargs):
        for d in data.parse(**kwargs):

            if not self.dict_keys:
                points = d
            else:
                points = d[self.dict_keys[0]]
                for k in range(1, len(self.dict_keys)):
                    try:
                        points = points[self.dict_keys[k]]
                    except KeyError as e:
                        return None

            if isinstance(points, list):
                for point in points:
                    self._enrich_point(point)
            else:
                self._enrich_point(points)

            yield d

    def enrich_geopandas_data(self, data):
        data = data.set_crs(epsg=4326).to_crs(epsg=3857)
        data["geometry"] = data.buffer(self.radius)
        data = data.to_crs(epsg=4326)
        self._df = self._df.set_crs(epsg=4326)

        spatial_joined = geopandas.sjoin(
            data, self._df, how="left", predicate="intersects"
        )
        return spatial_joined

    def enrich_dask_geopandas_data(self, data):
        data = data.set_crs("EPSG:4326").to_crs("EPSG:3857")
        data["geometry"] = data.buffer(self.radius)
        data = data.to_crs("EPSG:4326")
        self._df = self._df.set_crs("EPSG:4326")

        joined = dask_geopandas.sjoin(data, self._df, predicate="intersects")
        return joined

    def enrich(self, data, **kwargs):
        """Method overrided of interface. This method do enrichment using OSM data as a enricher. It walk through the keys to reach at the data that will be used to intersect the polygons. It uses a R tree to index polygons and search faster. If the radius attribute is passed the algorithm returns all polygons that intersect the point buffered with this radius else the algorithm returns all polygons that contains the point.

        Parameters
        ----------
        data: :obj:`Data`
        """

        if self.files is None and self.key is not None and self.value is not None:
            osm_util = OSM_util()
            self._df = osm_util.get_places(self.place_name, self.key, self.value)

        self._get_polygons()

        if isinstance(data, GeopandasData):
            return self.enrich_geopandas_data(data.data)

        elif isinstance(data, DaskGeopandasData):
            return self.enrich_dask_geopandas_data(data.data)

        elif isinstance(data, JsonData):
            return self.enrich_json_data(data, **kwargs)



from setuptools import setup, find_packages
import os

# get all data dirs in the datasets module
data_files = []

for item in os.listdir("moredata/datasets"):
    print('')
    if not item.startswith("__"):
        if os.path.isdir(os.path.join("moredata/datasets/", item)):
            data_files.append(os.path.join("datasets", item, "*"))

if os.environ.get("READTHEDOCS", False) == "True":
    INSTALL_REQUIRES = []
    DOCS_REQUIRES = []
else: 
    INSTALL_REQUIRES = [
        "dask-geopandas>=0.2.0",
        "requests>=2.22.0",
        "SQLAlchemy>=1.4.37",
        "PyMySQL>=0.10.0",
        "osm2geojson>=0.1.27",
        "h3==3.4.3",
        "Rtree>=0.9.7",
        "pyproj>=2.4.2.post1",
        "geopandas>=0.7.0",
        "geopy>=2.0.0",
        "pandas>=1.0.1",
        "numpy>=1.18.1",
        "elasticsearch==7.17.0",
        "shapely>=1.7.0",
    ]

    DOCS_REQUIRES = [
        "sphinx==2.4.0",
        "sphinx-rtd-theme==0.4.3",
        "jinja2==2.11.1",
        "sphinxcontrib-napoleon==0.7",
    ]

setup(
    name="moredata",
    url='https://github.com/gegen07/more-data',
    download_url='https://github.com/gegen07/more-data/archive/v0.1.5.tar.gz',
    version='0.1.5',
    packages=find_packages(where="."),
    package_data={"moredata": data_files},
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    extras_require={
        "develop": INSTALL_REQUIRES + DOCS_REQUIRES,
        "docs": DOCS_REQUIRES,
    },
)


import json
from elasticsearch.exceptions import TransportError
from elasticsearch.helpers import bulk, streaming_bulk, reindex, scan
from elasticsearch.helpers.errors import BulkIndexError


class IndexHandler:
    """Index Handler control index actions in Elasticsearch.

    Parameters
    ----------
    client: :obj:`elasticsearch.Elasticsearch`
    index: str
        name of index
    doc_type: str
        name of document

    Attributes
    ----------
    client: :obj:`elasticsearch.Elasticsearch`
    index: str
        name of index
    doc_type: str
        name of document
    """

    def __init__(self, client, index, doc_type):
        self.client = client
        self.index = index
        self.doc_type = doc_type

    def create_index(self, mapping):
        """This method create index with mapping provided using
        elasticsearch-py package (https://elasticsearch-py.readthedocs.io/en/master/api.html#elasticsearch.client.IndicesClient.create)

        Parameters
        ----------
        mapping: :obj:`Json`
            mapping is a definitions of attributes index types (https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping.html)
        """
        try:
            self.client.indices.create(index=self.index, body=mapping)
        except TransportError as e:
            if e.error == "resource_already_exists_exception":
                pass
            else:
                raise

    def load_index(self, parser, streaming=None, **kwargs):
        """load_index method load documents to specified index in constructor.
        It can stream the load or just load all data passing an `iterable` as parameter.
        The difference between of these two methods is that if you want that your data assume specified
        type you have to stream but if you don't care about this you can simply bulk data that
        elasticsearch will infer the types of your document.

        Parameters
        ----------
        pasrser: Callable
            It's a function that yield the document which has data to index into elasticsearch.

        streaming: bool
            If streaming load_index method will do streaming bulk instead of bulk all data.
        """
        try:
            if streaming:
                for ok, _ in streaming_bulk(
                    self.client, index=self.index, actions=parser(**kwargs)
                ):
                    if not ok:
                        raise Exception
            else:
                bulk(
                    self.client,
                    parser(**kwargs),
                    index=self.index,
                    doc_type=self.doc_type,
                )
        except BulkIndexError as e:
            pass

    def _remove_fields_preprocessed(self, data, **kwargs):
        array_point_field = kwargs.get("array_point_field")
        geo_location = kwargs.get("geo_location")
        code_h3 = kwargs.get("code_h3")

        if geo_location:
            if array_point_field:
                for points in data[array_point_field]:
                    points = points.pop("geo_location", None)
            else:
                data = data.pop("geo_location", None)

        if code_h3:
            if array_point_field:
                for points in data[array_point_field]:
                    points = points.pop("code_h3", None)
            else:
                data = data.pop("code_h3", None)

        return data

    def get_all_data(self, index, **kwargs):
        """Get all data indexed by a index.

        Parameters
        ----------
        index: str
            name of index

        query: dict
            query is the document that elasticsearch uses to retrive information

        Yields
        ------
        generate the data retrieved by elasticsearch.

        """

        query = {"query": {"match_all": {}}}

        for record in scan(self.client, query=query, index=index):
            if kwargs.get("geo_location") or kwargs.get("code_h3"):
                yield self._remove_fields_preprocessed(record["_source"], **kwargs)
            else:
                yield record["_source"]

    def re_index(self, reindex_handler):
        """reindex to apply pipeline to enrich

        Parameters
        ----------
        reindex_handler: :obj:`ReindexHandler`
        """
        self.client.reindex(reindex_handler._json)


class ReindexHandler:
    """ReindexHandler creates a necessary json to send to elasticsearch to reindex with pipeline
    and enrich the index provided.

    Parameters
    ----------
    index: str
        name of source index

    target_index: str
        name of destination index

    pipeline_name: str
        name of pipeline


    Attributes
    ----------

    json: dict
        this is a json file created by the parameters to sendo to reindex route
    """

    def __init__(self, index, target_index, pipeline_name):
        self._json = {
            "source": {"index": index},
            "dest": {"index": target_index, "pipeline": pipeline_name},
        }

        self.source_index = index
        self.target_index = target_index



from geopy.geocoders import Nominatim
from shapely.ops import polygonize
from shapely import geometry
import pandas as pd
import numpy as np
import geopandas as gpd
from shapely.ops import linemerge
import requests
import json
import osm2geojson


class OSM_util:
    def get_place_ID(self, place_name):
        geolocator = Nominatim(user_agent="city_compare")
        geoResults = geolocator.geocode(
            place_name, exactly_one=False, limit=3, timeout=600
        )
        city = None
        for r in geoResults:
            if r.raw.get("osm_type") == "relation":
                city = r
                break
        if city is None:
            raise Exception("city not found")
        place_ID = (
            int(city.raw.get("osm_id")) + 3600000000
        )  # Calcula o ID do local escolhido utilizando\n",
        return place_ID

    def _get_places_overpy(self, query):
        try:
            result = requests.get(
                "http://overpass-api.de/api/interpreter", data={"data": query}
            ).json()
            return osm2geojson.json2geojson(result)
        except:
            return {"features": [], "type": "FeatureCollection"}

    def get_places(self, place_name, key, value, query=None, tags=("name", "geom")):
        if query is None:
            place_id = self.get_place_ID(place_name)
            query = """
            [out:json][timeout:3600];
            area(%s)->.searchArea;
            (
                way[%s=%s](area.searchArea);
                relation[%s=%s](area.searchArea);
            );
            out geom;
            """ % (
                place_id,
                key,
                value,
                key,
                value,
            )

        result = self._get_places_overpy(query)

        for data in result["features"]:
            properties = data["properties"]
            if "tags" in properties.keys():
                tags = properties["tags"]
                properties["value"] = value
                properties["key"] = key
                for item in tags.items():
                    k, v = item[0], item[1]
                    properties[k] = v
        df = gpd.GeoDataFrame.from_features(result["features"], crs="EPSG:4326")
        return df



from ..enricher import IEnricherConnector
from ...models.data import Data
from .pipeline_handler import Pipeline, PipelineHandler
from .index_handler import ReindexHandler


class ElasticsearchConnector(IEnricherConnector):
    """ElasticsearchConnector implements interface IEnricherConnector,
    so this is a connector that can be used to enrich some data.
    ElasticsearchConnector implements handlers for `index`, `pipeline` and `policy`,
    all are requirement to enrich your data using elasticsearch.

    Parameters
    ----------
    index_handler: :obj:`IndexHandler`
        control index actions in elasticsearch

    pipeline: :obj:`Pipeline`
        control pipeline actions in elasticsearch

    reindex_handler: :obj:`ReindexHandler`
        control reindexing in elasticsearch

    Attributes
    ----------
    index_handler: :obj:`IndexHandler`

    pipeline: :obj:`Pipeline`

    reindex_handler: :obj:`ReindexHandler`
    """

    def __init__(self, index_handler, pipeline, reindex_handler):
        self.index_handler = index_handler
        self.pipeline = pipeline
        self.reindex_handler = reindex_handler

    def enrich(self, data, **kwargs):
        """Method overrided of interface. This interface do enrichment using
        elasticsearch as a enricher: create a pipeline, reindex with a pipeline
        specified to enrich and return all data enriched as Json.

        Parameters
        ----------
        data: :obj:`Data`
        """

        self.pipeline.create_pipeline()
        self.index_handler.re_index(self.reindex_handler)

        return self.index_handler.get_all_data(
            index=self.reindex_handler.target_index, **kwargs
        )
